﻿package {
	/**
	 * Библиотека для удобой подгрузки ресурсов. 
	 * Позволяет абстрагироваться от типа данных, указав URL и получив объект
	 * @author fr0z3n
	 * @author Yarick
	 * @version 2.0   
	 * @copy Классы Library.as и DataLoader.as являются свободно распростроняемыми (OpenSource)
	 * Релиз 21 июня 2011
	 */
	/**
	* Краткая документация по использованию:
	var lib:Library=new Library();//Создаём объект
	lib.add("www.site.ru/img.png",BitMap);  lib.add("model.xml",XML);   //Добавляем любое количество внешних URL на ресурсы
	//!Адрес должен быть без протокола и заканчиваться на расширение. В дальнейшем он будет идентифицировать ресурсы
	lib.addEventListener(Event.COMPLETE,func);//Добавить функцию-приёмник события конца загрузки
	lib.load();//Стартовать загрузку
	//Во время загрузки можно узнавать размер загруженной информации через  bytesLoaded
	//Загрузка закончиться после события COMPLETE
	//После окончания загрузки данные доступны через lib.data["www.site.ru/img.png"] - возвращает объект соответствуюшего класса
	//Также можно, указав convertNotTransparentIMG=true , добиться преобразования всех непрозрачных изображений в прозрачные
	*/
	import DataLoader;//Вспомогательный класс. Методом композиции (Не наследования) объединяет
	//URLLoader и Loader. Выбирает необходимый класс, основываясь на типе загружаемого объекта
	
	//Поддерживаемые типы данных
	import flash.display.Bitmap;
	import flash.display.MovieClip;
	import flash.display.Sprite;	
	//Для загрузки
	import flash.net.URLRequest;
	import flash.display.Loader;
	import flash.net.URLLoader;
	import flash.display.LoaderInfo;
	//Для работы с изображениями
	import flash.display.BitmapData;
	import flash.geom.Point;
	import flash.geom.Rectangle;

	import flash.events.*;
	
	public class Library extends EventDispatcher{
		private var loaders:Array = new Array();//массив загрузчиков для объектов
		private var isDone:Boolean = false;//Флаг окончания загрузки
		
		public var data:Object = new Object();//"Словарь" - по ключу получаем объект
		
		private var convertNotTransparentIMG:Boolean = false;//Преобразовывать непрозрачные изображения в прозрачные
		//На основе левого верхнего угла - он должен задавать цвет, который нужно заменить на альфа канал
		
		/**
		 * Создаём библиотеку
		 * @param	convertNotTransparentIMG нужно ли приобразовывать 
		 */
		public function Library(convertNotTransparentIMG:Boolean = false):void {
			this.convertNotTransparentIMG = convertNotTransparentIMG;//Запоминаем опцию
		}
		
		/**
		 * Добавть данные для загрузки
		 * @param	url адрес данных
		 * @param	_class класс, к которому требуется привести данные после загрузки
		 * Возвращает загрузчик
 		 */
		public function add(url:String, _class:Class):DataLoader {
			if (isDublicate(url)) { 
				return null; //Если ресурс - дубликат, то мы его не добавляем
			}
			
			var tempLoader:DataLoader = new DataLoader();//Создаём загрузчег
			tempLoader._class = _class;//Присваиваем класс
			
			//Выбираем тип данных в зависимости от класса
			switch(_class){
				case Bitmap:
				case Sprite:
				case MovieClip:
				tempLoader.loader = new Loader();
				tempLoader.loader.contentLoaderInfo.addEventListener(ProgressEvent.PROGRESS,progressListener);//При получении байт данных
				tempLoader.loader.contentLoaderInfo.addEventListener(IOErrorEvent.IO_ERROR, errorListener);//Обработчик ошибки			
				tempLoader.loader.contentLoaderInfo.addEventListener(Event.COMPLETE,completeListener);//При окончании загрузки одного из элементов
				break;
				case XML:
				tempLoader.urlLoader = new URLLoader();
				tempLoader.urlLoader.addEventListener(ProgressEvent.PROGRESS,progressListener);//При получении байт данных
				tempLoader.urlLoader.addEventListener(IOErrorEvent.IO_ERROR, errorListener);//Обработчик ошибки
				tempLoader.urlLoader.addEventListener(Event.COMPLETE,completeListener);//При окончании загрузки одного из элементов
				break;
				default:
				//Если класс не поддерживаеться - вылетаем с ошибкой
				throw new Error("Класс " + _class + " для " + url + " не поддерживаеться библиотекой");
			}
			
			tempLoader.url = url;//Запоминаем URL. он же будет ID объекта
		
			loaders.push(tempLoader);//Добавляем лоадер в массив для загрузки
			return(tempLoader);
		}
		
		/**
		 * Проверить, добавлен ли в библиотеку объект с таким URL
		 * @param	url
		 * @return
		 */
		public function isDublicate(url:String):Boolean{
			for each(var loader:DataLoader in loaders) {
				if (loader.url == url) {
					return true;
				}
			}
			return false;
		}

		/**
		 * Запустить загрузку всех ресурсов
		 */
		public function load():void
		{
			for each(var loader:DataLoader in loaders){//Для каждого лоадера
				//Стартуем загрузку
				switch(loader._class) {
					case Sprite:
					case MovieClip:
					case Bitmap:
						loader.loader.load(new URLRequest(loader.url));
					break;
					case XML:
						loader.urlLoader.load(new URLRequest(loader.url));
					break;
				}
			}
		}
		
		/**
		 * Выполнить с каждым объектом определённые действия
		 * @param	func
		 */
		public function forEach(func:Function):void{
			for each(var loader:DataLoader in loaders){//Для каждого лоадера
				switch(loader._class) {				
					case Bitmap:
						func(Bitmap(loader.loader.content));
					break;
					case XML:
						func(XML(loader.urlLoader.data),loader);
					break;
				}
			}
		}
		
		/**
		 * Оработчик ошибки
		 */
		private function errorListener(e:IOErrorEvent):void {
			dispatchEvent(new Event(ErrorEvent.ERROR));//Диспетчеризуем ошибку
		}
		
		/**
		 * При получении байтов
		 */
		private function progressListener(e:Event):void { 
			dispatchEvent(new Event(ProgressEvent.PROGRESS));
		}
		
		/**
		 * Сделать битмапу альфа-канал из цвета в Правом нижнем углу
		 * Функция нужна для того, чтобы добиться прозрачности у изображений, 
		 * основываясь на
		 * @param	bmp
		 */
		public function createAlpha(bmp:Bitmap):void {
			var bmpData:BitmapData = bmp.bitmapData;
			var zone:Rectangle = new Rectangle(0, 0, bmp.width, bmp.height);
			var result:BitmapData = new BitmapData(bmp.width,bmp.height);//Создаём новый битмап с прозрачностью
			var color:uint = bmp.bitmapData.getPixel32(0, 0);//Извлекаем заменяемый цвет
			bmpData.lock();//Для ускорения
			//Заменяем цвет color на 0x00000000, т.е. на прозрачный
			result.threshold(bmpData, zone, new Point(0,0), "==",color, 0x00000000,0xFFFFFFFF,true);
			bmpData.unlock();
			bmp.bitmapData.dispose();//Удаляем старый битмап
			bmp.bitmapData = result;
		}
		
		/**
		 * Слушатель окончания закрузки
		 */
		private function completeListener(e:Event):void { 
			var allLoaded:Boolean = true;//Предположение, что все загружены
			
			//Проходимся по всем объектам и проверяем, всё ли загружено
			for each (var loader:DataLoader in loaders) {
				switch(loader._class) {
					case Sprite:
					case MovieClip:
					case Bitmap:
						//Если хотя бы один объект не начал загружаться (мы не знаем его bytesTotal) или не до конца загрузился - мы не закончили 
						if ((loader.loader.contentLoaderInfo.bytesTotal == 0) || (loader.loader.contentLoaderInfo.bytesLoaded != loader.loader.contentLoaderInfo.bytesTotal)) {
							allLoaded = false;
					    }
					break;
					case XML:
						//Аналогично ^
						if ((loader.urlLoader.bytesTotal == 0) || (loader.urlLoader.bytesLoaded != loader.urlLoader.bytesTotal)) {
							allLoaded = false;
						}
					break;
				}
			}
			
			if ((!isDone) && (allLoaded)) {//Если действительно все загружены и мы первый раз это диспетчеризуем
				insertObjectsInData();//Выводим объекты в Data
				isDone = true;//Помечаем, что диспетчеризация COMPLETE была проведена
				//Возможно возникновение такой ситуации, когда все объекты быстро загрузились (например, с диска)
				//И при вызове completeListener несколько раз оказываеться, что bytesLoaded=bytesTotal
				//Поэтому мы помечаем isDone и диспетчеризируем только 1 раз
				dispatchEvent(new Event(Event.COMPLETE));
			}
		}
		
		/**
		 * Вытащить объекты из лоадеров и вставить в data
		 * Теперь они доступны как data[id объекта]
		 */
		public function insertObjectsInData():void{
			for each (var loader:DataLoader in loaders) {
				switch(loader._class) {
					case Sprite:
						data[loader.url] = Sprite(loader.loader.content);
					case MovieClip:
						data[loader.url] = MovieClip(loader.loader.content);
					break;
					case Bitmap:
						var bmp:Bitmap=Bitmap(loader.loader.content);//Получаем битмап
						
						//Если нам нужно преобразовать изображение из непрозрачного в прозрачное
						if ((convertNotTransparentIMG)&&(!bmp.bitmapData.transparent)){
							createAlpha(bmp);
						}
						
						data[loader.url] = bmp;
					break;
					case XML:
						data[loader.url] = XML(loader.urlLoader.data);
					break;
				}
			}
		}
		
		/**
		 * Получить число загруженых байтов
		 */
		public function get bytesLoaded():uint{
			var tempByte:uint = 0;//Счётчик
			for each (var element:DataLoader in loaders) {
				switch(element._class) {
					case Bitmap: 
					case MovieClip: 					
					case Sprite: {
						tempByte +=  element.loader.contentLoaderInfo.bytesLoaded;	
						break;
					}
					case XML: {
						tempByte +=  element.urlLoader.bytesLoaded;
						break;
					}					
				}
			}
			return tempByte;
		}
		
		/**
		 * Получить общее число байтов
		 */
		public function get bytesTotal():int{
			var tempByte:int = 0;//Счётчик
			for each (var element:DataLoader in loaders) {
				switch(element._class) {
					case Bitmap: 
					case MovieClip: 					
					case Sprite: {
						tempByte +=  element.loader.contentLoaderInfo.bytesTotal;	
						break;
					}
					case XML: {
						tempByte +=  element.urlLoader.bytesTotal;
						break;
					}					
				}
			}
			return tempByte;
		}
	}
}